@@ -4,7 +4,7 @@
     This file is part of Corrade.
 
     Copyright © 2007, 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015, 2016,
-                2017, 2018 Vladimír Vondruš <mosra@centrum.cz>
+                2017, 2018, 2019 Vladimír Vondruš <mosra@centrum.cz>
 
     Permission is hereby granted, free of charge, to any person obtaining a
     copy of this software and associated documentation files (the "Software"),
@@ -350,20 +350,29 @@ Arguments:
 -   `--benchmark-red N` --- deviation threshold for marking benchmark red
     (environment: `CORRADE_BENCHMARK_RED`, default: `0.25`)
 
-@section TestSuite-Tester-cmake CMake support and platform-specific goodies
+@section TestSuite-Tester-running Compiling and running tests
 
-While the test executables can be created in any way you want, there's also an
-@ref corrade-cmake-add-test "corrade_add_test()" CMake macro that creates the
-executable, links `Corrade::TestSuite` library to it and adds it to CTest.
-Besides that it is able to link other arbitrary libraries to the executable
-and specify a list of files that the tests used. It provides additional useful
-features on various platforms:
+In general, just compiling the executable and linking it to the TestSuite
+library is enough, no further setup is needed. When running, the test produces
+output to standard output / standard error and exits with non-zero code in case
+of a test failure.
+
+@subsection TestSuite-Tester-running-cmake Using CMake
+
+If you are using CMake, there's a convenience @ref corrade-cmake-add-test "corrade_add_test()"
+CMake macro that creates the executable, links `Corrade::TestSuite` library to
+it and adds it to CTest. Besides that it is able to link other arbitrary
+libraries to the executable and specify a list of files that the tests used. It
+provides additional useful features on various platforms:
 
 -   If compiling for Emscripten, using @ref corrade-cmake-add-test "corrade_add_test()"
     makes CTest run the resulting `*.js` file via Node.js. Also it is able to
     bundle all files specified in `FILES` into the virtual Emscripten
     filesystem, making it easy to run file-based tests on this platform; all
-    environment options are passed through as well.
+    environment options are passed through as well. The macro also creates a
+    runner for manual testing in a browser, see
+    @ref TestSuite-Tester-running-emscripten-browser "below" for more
+    information.
 -   If Xcode projects are generated via CMake and @ref CORRADE_TESTSUITE_TARGET_XCTEST
     is enabled, @ref corrade-cmake-add-test "corrade_add_test()" makes the test
     executables in a way compatible with XCTest, making it easy to run them
@@ -380,6 +389,59 @@ Emscripten and Android in path specified in `JPEG_TEST_DIR` that was saved into
 the `configure.h` file inside current build directory:
 
 @snippet testsuite.cmake 0
+
+@subsection TestSuite-Tester-running-android Manually running the tests on Android
+
+If not using CMake CTest, Android tests can be run manually. When you have
+developer-enabled Android device connected or Android emulator running, you can
+use ADB to upload the built test to device temp directory and run it there:
+
+@code{.sh}
+adb push <path-to-the-test-build>/MyTest /data/local/tmp
+adb shell /data/local/tmp/MyTest
+@endcode
+
+You can also use @cb{.sh} adb shell @ce to log directly into the device shell
+and continue from there. All @ref TestSuite-Tester-command-line "command-line arguments"
+are supported.
+
+@attention Keep in mind that older versions of ADB and Android do not correctly
+    propagate the exit code to caller, which may result in your test failures
+    being silently ignored. See [Android Issue 3254](http://web.archive.org/web/20160806094132/https://code.google.com/p/android/issues/detail?id=3254)
+    for possible workarounds. The @ref corrade-cmake-add-test "corrade_add_test()"
+    CMake macro also works around this issue.
+
+@subsection TestSuite-Tester-running-emscripten Manually running the tests on Emscripten
+
+When not using CMake CTest, Emscripten tests can be run directly using Node.js.
+Emscripten sideloads the WebAssembly or asm.js binary files from current
+working directory, so it's needed to @cb{.sh} cd @ce into the test build
+directory first:
+
+@code{.sh}
+cd <test-build-directory>
+node MyTest.js
+@endcode
+
+See also the `--embed-files` [emcc option](https://kripken.github.io/emscripten-site/docs/porting/files/packaging_files.html)
+for a possibility to bundle test files with the executable.
+
+@subsection TestSuite-Tester-running-emscripten-browser Running Emscripten tests in a browser
+
+Besides running tests using Node.js, it's possible to run each test case
+manually in a browser. Browsers require the executables to be accessed via a
+webserver --- if you have Python installed, you can simply start serving the
+contents of your build directory using the following command:
+
+@code{.sh}
+cd <test-build-directory>
+python -m http.server
+@endcode
+
+The webserver is then available at http://localhost:8000. It supports directory
+listing, so you can navigate to each test case runner HTML file (look for e.g.
+`MyTest.html`). Unfortunately it's at the moment not possible to run all
+browser tests in a batch or automate the process in any other way.
 */
 class CORRADE_TESTSUITE_EXPORT Tester {
     public:
@@ -417,15 +479,9 @@ class CORRADE_TESTSUITE_EXPORT Tester {
          * @brief Alias for debug output
          *
          * For convenient debug output inside test cases (instead of using
-         * fully qualified name):
+         * the fully qualified name):
          *
-         * @code{.cpp}
-         * void myTestCase() {
-         *     int a = 4;
-         *     Debug() << a;
-         *     CORRADE_COMPARE(a + a, 8);
-         * }
-         * @endcode
+         * @snippet TestSuite.cpp Tester-Debug
          *
          * @see @ref Warning, @ref Error
          */
@@ -469,13 +525,6 @@ class CORRADE_TESTSUITE_EXPORT Tester {
              */
             WallTime = 2,
 
-            #ifdef CORRADE_BUILD_DEPRECATED
-            /** @brief @copybrief BenchmarkType::WallTime
-             * @deprecated Use @ref BenchmarkType::WallTime instead.
-             */
-            WallClock CORRADE_DEPRECATED_ENUM("use BenchmarkType::WallTime instead") = int(WallTime),
-            #endif
-
             /**
              * CPU time. Suitable for measuring most events (microseconds and
              * up). While the reported time is in nanoseconds, the actual timer
@@ -514,25 +563,9 @@ class CORRADE_TESTSUITE_EXPORT Tester {
             /* Values should not overlap with BenchmarkType */
 
             Nanoseconds = 100,      /**< Time in nanoseconds */
-
-            #ifdef CORRADE_BUILD_DEPRECATED
-            /** @brief @copybrief BenchmarkUnits::Nanoseconds
-             * @deprecated Use @ref BenchmarkUnits::Nanoseconds instead.
-             */
-            Time CORRADE_DEPRECATED_ENUM("use Nanoseconds instead") = int(Nanoseconds),
-            #endif
-
             Cycles = 101,           /**< Processor cycle count */
             Instructions = 102,     /**< Processor instruction count */
             Bytes = 103,            /**< Memory (in bytes) */
-
-            #ifdef CORRADE_BUILD_DEPRECATED
-            /** @brief @copybrief BenchmarkUnits::Bytes
-             * @deprecated Use @ref BenchmarkUnits::Bytes instead.
-             */
-            Memory CORRADE_DEPRECATED_ENUM("use Bytes instead") = int(Bytes),
-            #endif
-
             Count = 104             /**< Generic count */
         };
 
@@ -904,9 +937,7 @@ class CORRADE_TESTSUITE_EXPORT Tester {
          * By default the test case name is gathered in the check macros and is
          * equivalent to the following:
          *
-         * @code{.cpp}
-         * setTestCaseName(__func__);
-         * @endcode
+         * @snippet TestSuite.cpp Tester-setTestCaseName
          *
          * @see @ref setTestName(), @ref setTestCaseDescription()
          */
@@ -982,7 +1013,10 @@ class CORRADE_TESTSUITE_EXPORT Tester {
         }
 
         /* Compare two different types with explicit comparator specification */
-        template<class T, class U, class V> void compareWith(Comparator<T> comparator, const std::string& actual, const U& actualValue, const std::string& expected, const V& expectedValue);
+        template<class T, class U, class V> void compareWith(Comparator<T>& comparator, const std::string& actual, const U& actualValue, const std::string& expected, const V& expectedValue);
+        template<class T, class U, class V> void compareWith(Comparator<T>&& comparator, const std::string& actual, const U& actualValue, const std::string& expected, const V& expectedValue) {
+            return compareWith<T, U, V>(comparator, actual, actualValue, expected, expectedValue);
+        }
 
         template<class T> void verify(const std::string& expression, T&& value);
 
@@ -1093,14 +1127,15 @@ class CORRADE_TESTSUITE_EXPORT Tester {
         std::vector<TestCase> _testCases;
         std::string _testFilename, _testName, _testCaseName,
             _testCaseDescription, _benchmarkName, _expectFailMessage;
-        std::size_t _testCaseId, _testCaseInstanceId, _testCaseRepeatId,
-            _benchmarkBatchSize, _testCaseLine, _checkCount;
+        std::size_t _testCaseId{}, _testCaseInstanceId{~std::size_t{}},
+            _testCaseRepeatId{}, _benchmarkBatchSize{}, _testCaseLine{},
+            _checkCount{};
 
-        std::uint64_t _benchmarkBegin;
-        std::uint64_t _benchmarkResult;
+        std::uint64_t _benchmarkBegin{};
+        std::uint64_t _benchmarkResult{};
         TestCase* _testCase = nullptr;
-        bool _expectedFailuresDisabled;
-        ExpectedFailure* _expectedFailure;
+        bool _expectedFailuresDisabled{};
+        ExpectedFailure* _expectedFailure{};
         TesterConfiguration _configuration;
 };
 
@@ -1152,19 +1187,13 @@ namespace.
 If the expression is not true, the expression is printed and execution of given
 test case is terminated. Example usage:
 
-@code{.cpp}
-string s("hello");
-CORRADE_VERIFY(!s.empty());
-@endcode
+@snippet TestSuite.cpp CORRADE_VERIFY
 
 It is possible to use @ref CORRADE_VERIFY() also on objects with
 @cpp explicit operator bool() @ce without doing explicit conversion (e.g. using
 @cpp !! @ce), for example:
 
-@code{.cpp}
-std::unique_ptr<T> t(new T);
-CORRADE_VERIFY(t);
-@endcode
+@snippet TestSuite.cpp CORRADE_VERIFY-explicit
 
 @see @ref CORRADE_COMPARE(), @ref CORRADE_COMPARE_AS()
 */
@@ -1180,10 +1209,7 @@ CORRADE_VERIFY(t);
 If the values are not the same, they are printed for comparison and execution
 of given test case is terminated. Example usage:
 
-@code{.cpp}
-int a = 5 + 3;
-CORRADE_COMPARE(a, 8);
-@endcode
+@snippet TestSuite.cpp CORRADE_COMPARE
 
 Comparison of floating-point types is by default done as a fuzzy-compare, see
 @ref Corrade::TestSuite::Comparator<float> "TestSuite::Comparator<float>" and
@@ -1206,9 +1232,7 @@ comparison operators and is printable via @ref Corrade::Utility::Debug "Utility:
 If the values are not the same, they are printed for comparison and execution
 of given test case is terminated. Example usage:
 
-@code{.cpp}
-CORRADE_COMPARE_AS(std::sin(0.0), 0.0f, float);
-@endcode
+@snippet TestSuite.cpp CORRADE_COMPARE_AS
 
 See also @ref Corrade::TestSuite::Comparator "TestSuite::Comparator" class documentation
 for example of more involved comparisons.
@@ -1234,9 +1258,7 @@ with given comparator type.
 If the values are not the same, they are printed for comparison and execution
 of given test case is terminated. Example usage:
 
-@code{.cpp}
-CORRADE_COMPARE_WITH("actual.txt", "expected.txt", Compare::File("/common/path/prefix"));
-@endcode
+@snippet TestSuite.cpp CORRADE_COMPARE_WITH
 
 See @ref Corrade::TestSuite::Comparator "TestSuite::Comparator" class
 documentation for more information.
@@ -1262,15 +1284,7 @@ and @ref CORRADE_COMPARE_AS() checks in the same scope. In most cases it will
 be until the end of the function, but you can limit the scope by placing
 relevant checks in a separate block:
 
-@code{.cpp}
-{
-    CORRADE_EXPECT_FAIL("Not implemented.");
-    CORRADE_VERIFY(isFutureClear());
-}
-
-int i = 6*7;
-CORRADE_COMPARE(i, 42);
-@endcode
+@snippet TestSuite.cpp CORRADE_EXPECT_FAIL
 
 If any of the following checks passes, an error will be printed to output.
 @see @ref CORRADE_EXPECT_FAIL_IF()
@@ -1289,24 +1303,11 @@ With @ref CORRADE_EXPECT_FAIL() it's not possible to write code such as this,
 because the scope of expected failure will end at the end of the @cpp if @ce
 block:
 
-@code{.cpp}
-{
-    if(answer != 42)
-        CORRADE_EXPECT_FAIL("This is not our universe.");
-
-    CORRADE_VERIFY(6*7, 49); // always fails
-}
-@endcode
+@snippet TestSuite.cpp CORRADE_EXPECT_FAIL_IF-wrong
 
 The solution is to use @cpp CORRADE_EXPECT_FAIL_IF() @ce:
 
-@code{.cpp}
-{
-    CORRADE_EXPECT_FAIL_IF(answer != 42, "This is not our universe.");
-
-    CORRADE_VERIFY(6*7, 49); // expect the failure if answer is not 42
-}
-@endcode
+@snippet TestSuite.cpp CORRADE_EXPECT_FAIL_IF
 
 Similarly to @ref CORRADE_VERIFY(), it is possible to use
 @ref CORRADE_EXPECT_FAIL_IF() also on objects with @cpp explicit operator bool @ce
@@ -1323,11 +1324,7 @@ without doing explicit conversion (e.g. using @cpp !! @ce).
 Skips all following checks in given test case. Useful for e.g. indicating that
 given feature can't be tested on given platform:
 
-@code{.cpp}
-if(!bigEndian) {
-    CORRADE_SKIP("Big endian compatibility can't be tested on this system.");
-}
-@endcode
+@snippet TestSuite.cpp CORRADE_SKIP
 */
 #define CORRADE_SKIP(message)                                               \
     do {                                                                    \
@@ -1347,14 +1344,7 @@ optimizer from removing the benchmark code such as assigning to a
 @cpp volatile @ce variable or combining all the results to a variable, which is
 then being used outside of the loop.
 
-@code{.cpp}
-void benchmark() {
-    std::string a = "hello", b = "world";
-    CORRADE_BENCHMARK(1000) {
-        volatile std::string c = a + b;
-    }
-}
-@endcode
+@snippet TestSuite.cpp CORRADE_BENCHMARK
 
 The resulting measured value is divided by @p batchSize to represent cost of
 one iteration.
@@ -1363,7 +1353,7 @@ one iteration.
     _CORRADE_REGISTER_TEST_CASE();                                          \
     for(CORRADE_UNUSED auto&& _CORRADE_HELPER_PASTE(benchmarkIteration, __func__): Tester::createBenchmarkRunner(batchSize))
 
-template<class T, class U, class V> void Tester::compareWith(Comparator<T> comparator, const std::string& actual, const U& actualValue, const std::string& expected, const V& expectedValue) {
+template<class T, class U, class V> void Tester::compareWith(Comparator<T>& comparator, const std::string& actual, const U& actualValue, const std::string& expected, const V& expectedValue) {
     ++_checkCount;
 
     /* Store (references to) possibly implicitly-converted values,
